from chilero.pg import parse_pgurl


def test_parse_pgurl():
    pg0 = parse_pgurl(
        'psycopg2+postgresql://postgres@localhost/test'
    )

    assert pg0['user'] == 'postgres'
    assert pg0['password']== None
    assert pg0['database'] == 'test'
    assert pg0['host'] == 'localhost'
    assert pg0['port'] == 5432

    pg1 = parse_pgurl(
        'psycopg2+postgresql://postgres@localhost:5432/test'
    )

    assert pg1['user'] == 'postgres'
    assert pg1['password']== None
    assert pg1['database'] == 'test'
    assert pg1['host'] == 'localhost'
    assert pg1['port'] == 5432

    pg2 = parse_pgurl(
        'psycopg2+postgresql://other:secret@host:15432/testing'
    )

    assert pg2['user'] == 'other'
    assert pg2['password'] == 'secret'
    assert pg2['database'] == 'testing'
    assert pg2['host'] == 'host'
    assert pg2['port'] == 15432



import asyncio
import json
import os
from urllib.parse import quote_plus

from aiohttp.web_exceptions import HTTPBadRequest, HTTPConflict, HTTPNotFound
from chilero.web import Resource as BaseResource
from chilero.web import Response
from psycopg2._psycopg import DatabaseError


class Resource(BaseResource):
    list_query = 'SELECT * from {table}'
    object_query = 'SELECT * FROM {table}'
    table_name = None
    count_query = 'SELECT count(1) from {table}'
    order_by = 'id'
    id_column = 'id'
    allowed_fields = []
    required_fields = []
    search_fields = None

    @asyncio.coroutine
    def get_pool(self):
        return self.app.get_pool()

    @asyncio.coroutine
    def get_cursor(self):
        pool = yield from self.get_pool()
        cursor = yield from pool.cursor()
        return cursor

    def get_count_query(self):
        return self.count_query.format(table=self.get_table_name())

    def get_list_query(self):
        return self.list_query.format(table=self.get_table_name())

    def get_object_query(self):
        return self.object_query.format(table=self.get_table_name())

    def get_list_query_filters(self, conditions=None, search=None):
        query_filters = ''
        query_args = tuple()
        if conditions:
            fields = conditions.keys()
            filters = ', '.join(
                ['{field_name}=%s'.format(field_name=f) for f in fields]
            )
            query_filters = '{query} WHERE {filters}'.format(
                query=query_filters, filters=filters
            )

            query_args = tuple([conditions[f] for f in fields])

        if search and self.search_fields:
            search_keywords = search.split()
            filters = ' or '.join(
                '{} ILIKE  %s'.format(f)
                for f in self.search_fields for x in search_keywords
            )
            separator = 'WHERE' if 'WHERE' not in query_filters else 'AND'
            query_filters = '{query} {separator} {filters}'.format(
                query=query_filters, separator=separator, filters=filters
            )
            query_args += tuple(
                [
                    '%{}%'.format(kw) for kw in search_keywords
                    for x in self.search_fields
                ]
            )

        return query_filters, query_args

    @asyncio.coroutine
    def get_count(self, conditions=None, search=None):
        query_filters, query_args = \
            self.get_list_query_filters(conditions, search)
        count_query = self.get_count_query()
        if asyncio.iscoroutine(count_query):
            count_query = yield from count_query

        query = ' '.join([count_query, query_filters])

        with (yield from self.get_cursor()) as cur:
            yield from cur.execute(query, query_args)
            return (yield from cur.fetchone())[0]

    def index(self):
        index = yield from self.do_index()
        return self.response(index)

    @property
    def offset(self):
        return int(self.request.GET.get('offset') or 0)

    @property
    def limit(self):
        from_url = self.request.GET.get('limit')
        from_env = int(os.environ.get('PAGE_LIMIT') or 20)
        return int(from_url or from_env)

    @property
    def prev_offset(self):
        if self.limit <= 0:
            return 0

        offset = self.offset - self.limit
        return offset if offset > 0 else 0

    @classmethod
    def relation(cls, resource, **kwargs):
        base = os.getenv('BASE_URL', 'http://localhost:8000')
        return dict(
            tipo='relation',
            resource='{}/{}'.format(base, resource),
            **kwargs
        )

    @property
    def next_offset(self):
        if self.limit <= 0:
            return 0

        return self.offset + self.limit

    def _build_url(self, args):
        return '{}?{}'.format(
            self.get_index_url(),
            '&'.join(
                ['{}={}'.format(k, v) for k, v in args.items()]
            )
        )

    def next_url(self, conditions, count, search):

        if self.next_offset <= self.offset or self.next_offset >= count:
            return None

        args = conditions.copy()
        if search:
            args.update(dict(search=quote_plus(search)))

        args.update(
            dict(
                offset=self.next_offset,
                limit=self.limit
            )
        )
        return self._build_url(args)

    def prev_url(self, conditions, search):
        if self.prev_offset >= self.offset:
            return None

        args = conditions.copy()
        if search:
            args.update(dict(search=quote_plus(search)))

        args.update(
            dict(
                offset=self.prev_offset,
                limit=self.limit
            )
        )
        return self._build_url(args)

    def set_limit(self, query):
        if self.offset > 0:
            query = '{query} OFFSET {offset}'.format(
                query=query, offset=self.offset
            )
        return query

    def set_offset(self, query):
        if self.limit > 0:
            query = '{query} LIMIT {limit}'.format(
                query=query, limit=self.limit
            )

        return query

    def do_index(self, conditions=None):
        conditions = conditions or {}
        search = self.request.GET.get('search')
        query_filters, query_args = \
            self.get_list_query_filters(conditions, search)
        order_by = ' order by {}'.format(self.order_by)
        get_list_query = self.get_list_query()
        if asyncio.iscoroutine(get_list_query):
            get_list_query = yield from get_list_query
        query = ' '.join([get_list_query, query_filters, order_by])
        query = self.set_offset(self.set_limit(query))
        count = yield from self.get_count(conditions, search)
        response = dict(
            self=self.get_self_url(),
            data=dict(
                offset=self.offset,
                limit=self.limit,
                next=self.next_url(conditions, count, search),
                prev=self.prev_url(conditions, search),
                count=count
            ),
            index=[]
        )
        pool = yield from self.get_pool()
        with (yield from pool.cursor()) as cur:
            yield from cur.execute(query, query_args)
            for row in (yield from cur.fetchall()):
                row = self.before_serialization(row)
                if asyncio.iscoroutine(row):  # pragma: no cover
                    row = yield from row
                obj = self.serialize_object(row)
                if asyncio.iscoroutine(obj):  # pragma: no cover
                    obj = yield from obj
                obj = self.after_serialization(obj)
                if asyncio.iscoroutine(obj):
                    obj = yield from obj

                response['index'].append(obj)
        response['data']['length'] = len(response['index'])
        return response

    def serialize_list_object(self, row):
        return self.serialize_object(row)

    def serialize_object(self, row):  # pragma: no cover
        return row

    def before_serialization(self, row):
        return row

    def after_serialization(self, obj):
        return obj

    def show(self, id, **kwargs):
        pool = yield from self.get_pool()

        with (yield from pool.cursor()) as cur:
            yield from cur.execute(
                '{query} where {id_column} = %s'.format(
                    query=self.get_object_query(), id_column=self.id_column
                ), (id,)
            )
            record = yield from cur.fetchone()

            if record is None:
                raise HTTPNotFound()

            record = self.before_serialization(record)
            if asyncio.iscoroutine(record):
                record = yield from record
            obj = self.serialize_object(record)
            obj = self.after_serialization(obj)
            if asyncio.iscoroutine(obj):
                obj = yield from obj

            return self.response(obj)

    def get_table_name(self):
        if self.table_name:  # pragma: no cover
            return self.table_name
        return self.get_resource_name()

    @asyncio.coroutine
    def before_update(self, cursor):  # pragma: no cover
        return cursor

    @asyncio.coroutine
    def after_update(self, cursor):  # pragma: no cover
        return cursor

    def prepare_update(self, data):  # pragma: no cover
        return data

    def prepare_insert(self, data):  # pragma: no cover
        return data

    @asyncio.coroutine
    def before_insert(self, cursor):  # pragma: no cover
        return cursor

    @asyncio.coroutine
    def after_insert(self, cursor, id):  # pragma: no cover
        return cursor

    def get_allowed_fields(self):
        return self.allowed_fields

    def get_required_fields(self):
        return self.required_fields

    def error_response(self, message, **kwargs):
        kwargs['message'] = str(message)
        return json.dumps(kwargs, indent=4).encode('utf-8')

    def validate_allowed_fields(self, data):
        for f in data.keys():
            if f not in self.get_allowed_fields():
                raise HTTPBadRequest(
                    body=self.error_response(
                        'Field "{field_name}" is not allowed'.format(
                            field_name=f
                        )
                    )
                )

    def validate_required_fields(self, data):
        for f in self.get_required_fields():
            if f not in data.keys():
                raise HTTPBadRequest(
                    body=self.error_response(
                        'Field "{field_name}" is required'.format(
                            field_name=f
                        )
                    )
                )

    def update(self, id, **kwargs):
        data = yield from self.request.json()
        self.validate_allowed_fields(data)

        data = self.prepare_update(data)
        if asyncio.iscoroutine(data):  # pragma: no cover
            data = yield from data

        updated_fields = data.keys()

        pool = yield from self.get_pool()
        with(yield from pool.cursor()) as cur:
            yield from self.before_update(cur)
            query = 'update {table} set {fields} where {id_column}=%s'.format(
                table=self.get_table_name(),
                fields=','.join(['{}=%s'.format(x)for x in updated_fields]),
                id_column=self.id_column
            )
            try:
                yield from cur.execute(
                    query, tuple([data[f] for f in updated_fields]+[id])
                )
            except DatabaseError as e:
                raise HTTPConflict(
                    body=self.error_response(e)
                )
            yield from self.after_update(cur)

        return Response(status=204)

    def new(self, **kwargs):
        data = yield from self.request.json()
        if asyncio.iscoroutinefunction(self.validate_allowed_fields):
            yield from self.validate_allowed_fields(data)
        else:
            self.validate_allowed_fields(data)

        if asyncio.iscoroutinefunction(self.required_fields):
            yield from self.validate_required_fields(data)
        else:
            self.validate_required_fields(data)

        data = self.prepare_insert(data)
        if asyncio.iscoroutine(data):  # pragma: no cover
            data = yield from data

        fields = data.keys()

        pool = yield from self.get_pool()
        with (yield from pool.cursor()) as cur:
            query = (
                'INSERT INTO {table} ({fields}) '
                'VALUES ({values}) '
                'returning {id_column}'
            ).format(
                table=self.get_table_name(),
                fields=','.join(fields),
                values=','.join(['%s' for x in fields]),
                id_column=self.id_column
            )

            values = (data[x] for x in fields)

            yield from self.before_insert(cur)
            try:
                yield from cur.execute(query, tuple(values))
            except DatabaseError as e:
                raise HTTPConflict(
                    body=self.error_response(e)
                )
            record_id = (yield from cur.fetchone())[0]
            yield from self.after_insert(cur, record_id)

        return Response(
            status=201,
            headers=(('Location', self.get_object_url(record_id)),)
        )

    def destroy(self, **kwargs):
        pool = yield from self.get_pool()
        with(yield from pool.cursor()) as cur:
            query = 'DELETE FROM {table} where id={id_column}'.format(
                table=self.get_table_name(),
                id_column=kwargs['id']
            )
            try:
                yield from cur.execute(
                    query
                )
            except DatabaseError as e:
                raise HTTPConflict(
                    body=self.error_response(e)
                )

        return Response(status=200)



import asyncio

import aiopg
from chilero.pg.utils import parse_pgurl


@asyncio.coroutine
def get_pool(**kwargs):  # pragma: no cover
    '''
    Gets the postgresql connection pool.

    :return: pool
    '''
    db_url = kwargs.pop('db_url')
    kwargs.update(parse_pgurl(db_url))

    pool = yield from aiopg.create_pool(**kwargs)

    return pool



import asyncio
import random

import names
from chilero.web.test import asynctest

from chilero.pg import Resource
from chilero.pg.test import TestCase, TEST_DB_SUFFIX


class Friends(Resource):
    order_by = 'name ASC'
    search_fields = ['name']
    allowed_fields = ['name']

    def serialize_object(self, row):
        return dict(
            id=row[0],
            name=row[1],
            url=self.get_object_url(row[0])
        )


class BaseTestCase(TestCase):
    settings = dict(
        db_url='postgres://postgres@localhost:5432/chilero_pg_{}'.format(
            TEST_DB_SUFFIX
        )
    )
    routes = [
        ['/friends', Friends]
    ]

    @asyncio.coroutine
    def _create_friend(self, **kwargs):
        defaults = dict(
            name=self._random_string()
        )

        return(
            yield from self._create_and_get('/friends', kwargs, defaults)
        )


class TestAdvancedOptions(BaseTestCase):

    @asyncio.coroutine
    def _a_lot_of_friends(self):
        # create a lot of friends
        all_names = []
        for i in range(100):
            name = names.get_full_name()
            all_names.append(name)
            _, f = yield from self._create_friend(name=name)
            _.close()
        return all_names

    @asynctest
    def test_pagination(self):

        yield from self._a_lot_of_friends()

        # list with default values
        # page 1
        r = yield from self._get_json(self.full_url('/friends'))
        assert r['data']['count'] >= 100
        assert r['data']['prev'] == None
        assert 'offset=20' in r['data']['next']
        assert 'limit=20' in r['data']['next']
        assert len(r['index']) == r['data']['length']

        # page 2
        r = yield from self._get_json(r['data']['next'])
        assert 'offset=0' in r['data']['prev']
        assert 'offset=40' in r['data']['next']
        assert len(r['index']) == r['data']['length']

    @asynctest
    def test_pagination_no_limit(self):

        yield from self._a_lot_of_friends()

        # list with no limit
        r = yield from self._get_json(self.full_url('/friends?limit=0'))

        assert r['data']['count'] >= 100
        assert r['data']['prev'] == None
        assert r['data']['next'] == None
        assert r['data']['length'] == r['data']['count']
        assert len(r['index']) == r['data']['count']

    @asynctest
    def test_search_pagination(self):
        rnames = list((yield from self._a_lot_of_friends()))
        rname = random.choice(rnames).split()[0]
        for i in range(5):
            name = '{} {}'.format(rname, names.get_last_name())
            _, friend = yield from self._create_friend(name=name)
            _.close()

        rname = rname.lower()

        r = yield from self._get_json(
            self.full_url('/friends?search={}&limit=1'.format(rname))
        )

        assert r['data']['count'] >= 1
        assert rname in r['data']['next']

        while r['data']['next']:
            r = yield from self._get_json(r['data']['next'])
            if r['data']['next'] is not None:
                assert rname in r['data']['next']
            assert rname in r['data']['prev']
            rname.lower() in r['index'][0]['name'].lower()


class TestBasic(BaseTestCase):
    # Test common REST actions

    @asynctest
    def test_index(self):
        resp = yield from self._get(self.full_url('/friends'))

        assert resp.status == 200
        resp.close()

    @asynctest
    def test_index_json(self):
        resp = yield from self._index('/friends')
        assert isinstance(resp, dict)
        assert 'index' in resp

    @asynctest
    def test_create(self):
        name = self._random_string()
        _, friend = yield from self._create_friend(name=name)
        assert _.status == 201
        _.close()
        assert friend['name'] == name
        efriend = yield from self._delete(friend['url'])
        assert efriend.status==200

    @asynctest
    def test_create_error(self):
        _, friend = yield from self._create_friend(wrong_field=123)
        assert _.status == 400
        _.close()

    @asynctest
    def test_create_conflict(self):
        name = names.get_full_name()
        _, friend = yield from self._create_friend(name=name)
        _.close()
        _, friend = yield from self._create_friend(name=name)
        assert _.status == 409
        _.close()


    @asynctest
    def test_update(self):
        _, friend = yield from self._create_friend()
        _.close()

        new_name = self._random_string()

        presp = yield from self._patch(friend['url'], name=new_name)

        assert presp.status == 204
        presp.close()

        updated_friend = yield from self._get_json(friend['url'])
        assert updated_friend['body']['name'] == new_name

    @asynctest
    def test_search(self):
        name = 'some known name'
        _, friend = yield from self._create_friend(name=name)
        _.close()
        results = yield from self._search('/friends', terms='known name')

        assert len(results['index']) > 0
        assert results['index'][0]['name'] == name

    @asynctest
    def test_view_404(self):
        resp = yield from self._get(self.full_url('/friends/999999'))
        assert resp.status == 404
        resp.close()

    @asynctest
    def test_update_400(self):
        _, friend = yield from self._create_friend()
        _.close()

        new_name = self._random_string()

        presp = yield from self._patch(friend['url'], names=new_name)

        assert presp.status == 400
        presp.close()





import setuptools

setup_params = dict(
    name='chilero_pg',
    use_scm_version=True,
    namespace_packages=['chilero'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    url='https://github.com/dmonroy/chilero.pg',
    author='Darwin Monroy',
    author_email='contact@darwinmonroy.com',
    description='PostgreSQL utilities for chilero',
    install_requires=[
        'aiopg',
        'chilero>=0.3.7'
    ],
    setup_requires=[
        'setuptools_scm',
    ],
)


if __name__ == '__main__':
    setuptools.setup(**setup_params)
